
from typing import List, Dict, Any

from ccxt.base.types import Position

from exchanges.any_exchange import AnyExchange

'''
Example,
    spot        'BTC/USDT'
    perpetual   'BTC/USDT:USDT'
In both cases, BTC is base ccy, and USDT is quote ccy.

crypto/CCXT convention is, 
    For spot, in base ccy. 
    For perpetual contracts, this is in # contracts, not base ccy.
Here, Order.amount is always in base ccy regardless whether you're trading spot or perpetual. 

leg_room_bps: 
    For limit orders, when order is executed, limit price be ...
        buy order: best ask * (1 + leg_room_bps/10000) 
        sell order: best bid * (1 - leg_room_bps/10000)
    Thus a positive leg room means you're more aggressive to trying to get the order filled: Buy at higher price, Sell at lower price. 
'''
class Order:
    def __init__(
        self,
        ticker : str,
        side : str,  # buy/sell
        amount : float,
        order_type : str, # market/limit
        leg_room_bps : float = 0
    ) -> None:
        self.ticker = ticker
        self.side = side.strip().lower()
        self.amount = amount
        self.order_type = order_type.strip().lower()
        self.leg_room_bps = leg_room_bps

    def to_dict(self) -> Dict[Any, Any]:
        return {
            "ticker" : self.ticker,
            "side" : self.side,
            "amount" : self.amount,
            "order_type" : self.order_type,
            "leg_room_bps" : self.leg_room_bps
        }

'''
For limit orders, if not filled within wait_fill_threshold_ms, we'd try cancel and resend un-filled amount as market order.
wait_fill_threshold_ms default to -1: Means wait forever until fully filled.
'''
class DivisiblePosition(Order):
    def __init__(
        self,
        ticker : str,
        side : str,  # buy/sell
        amount : float,
        order_type : str, # market/limit
        leg_room_bps : float,
        slices : int = 1,
        wait_fill_threshold_ms : float = -1
    ) -> None:
        super().__init__(ticker, side, amount, order_type, leg_room_bps)
        self.slices = slices
        self.wait_fill_threshold_ms = wait_fill_threshold_ms
        self.multiplier = 1
        self.filled_amount = 0
        self.average_cost = 0

        self.executions : Dict[str, Dict[str, Any]] = {}

    def to_slices(self) -> List[Order]:
        slices : List[Order] = []

        remaining_amount_in_base_ccy : float = self.amount
        slice_amount_in_base_ccy : float = self.amount / self.slices
        for i in range(self.slices):
            if remaining_amount_in_base_ccy>0:
                if remaining_amount_in_base_ccy >= slice_amount_in_base_ccy:
                    slice = Order(
                        ticker=self.ticker, 
                        side=self.side, 
                        amount=slice_amount_in_base_ccy, 
                        leg_room_bps=self.leg_room_bps, 
                        order_type=self.order_type)
                    
                else:
                    # Last slice
                    slice = Order(
                        ticker=self.ticker, 
                        side=self.side, 
                        amount=remaining_amount_in_base_ccy, 
                        leg_room_bps=self.leg_room_bps, 
                        order_type=self.order_type)

                slices.append(slice)
                    
        return slices

    def append_execution(
        self,
        order_id : str,
        execution : Dict[str, Any]
    ):
        self.executions[order_id] = execution

    def get_execution(
        self,
        order_id : str
    ) -> Dict[str, Any]:
        return self.executions[order_id]

    def get_executions(self) -> Dict[str, Dict[str, Any]]:
        return self.executions

    def get_filled_amount(self) -> float:
        # filled_amount is in base ccy
        filled_amount = sum([ self.executions[order_id]['filled'] * self.multiplier for order_id in self.executions ])
        self.filled_amount = filled_amount
        return filled_amount

    def get_average_cost(self) -> float:
        average_cost = sum([ self.executions[order_id]['average'] * self.executions[order_id]['amount'] for order_id in self.executions ])
        average_cost = average_cost / sum([ self.executions[order_id]['amount'] for order_id in self.executions ])
        self.average_cost = average_cost
        return average_cost

    def to_dict(self) -> Dict[Any, Any]:
        rv = super().to_dict()
        rv['slices'] = self.slices
        rv['wait_fill_threshold_ms'] = self.wait_fill_threshold_ms
        rv['executions'] = self.executions
        rv['filled_amount'] = self.filled_amount
        rv['average_cost'] = self.average_cost
        return rv



from ccxt.base.exchange import Exchange

'''
The idea here is, if we were to trade thru for example IKBR, which is a tradfi broker not supported by CCXT, we would implement IBKR like CCXT implement for crypto exchanges.
Concrete implementation will:
a. Inherit from AnyExchange, thus ccxt Exchange class
b. Override ccxt basic functions
    - load_markets
    - fetch_balance
    - create_order
    - update_order
    - cancel_order
    - fetch_order (REST) vs watch_orders (websocket)
    - order amount rounding: amount_to_precision
    - order price rounding: price_to_precision
    ... etc
'''
class AnyExchange(Exception):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)



import os
import sys
import traceback
from enum import Enum
import argparse
from datetime import datetime, timedelta
import time
from typing import Dict, Union
import json
import asyncio
import logging
from ccxt import deribit
from redis import StrictRedis
import pandas as pd

from util.market_data_util import timestamp_to_datetime_cols, fetch_deribit_btc_option_expiries, fetch_ohlcv_one_candle

param : Dict = {
    'market' : 'BTC',

    # Provider ID is part of mds publish topic. 
    'provider_id' : 'b0f1b878-c281-43d7-870a-0347f90e6ece',

    'archive_file_name' : "deribit_options_expiry.csv",

    # Publish to message bus
    'mds' : {
        'topics' : {
            'deribit_options_expiry_publish_topic' : 'deribit-options-expiry'
        },
        'redis' : {
            'host' : 'localhost',
            'port' : 6379,
            'db' : 0,
            'ttl_ms' : 1000*60*15 # 15 min?
        }

    }    
}

logging.Formatter.converter = time.gmtime
logger = logging.getLogger()
log_level = logging.INFO # DEBUG --> INFO --> WARNING --> ERROR
logger.setLevel(log_level)
format_str = '%(asctime)s %(message)s'
formatter = logging.Formatter(format_str)
sh = logging.StreamHandler()
sh.setLevel(log_level)
sh.setFormatter(formatter)
logger.addHandler(sh)
# fh = logging.FileHandler(f"{param['job_name']}.log")
# fh.setLevel(log_level)
# fh.setFormatter(formatter)     
# logger.addHandler(fh)

class LogLevel(Enum):
    CRITICAL = 50
    ERROR = 40
    WARNING = 30
    INFO = 20
    DEBUG = 10
    NOTSET = 0

def log(message : str, log_level : LogLevel = LogLevel.INFO):
    if log_level.value<LogLevel.WARNING.value:
        logger.info(f"{datetime.now()} {message}")

    elif log_level.value==LogLevel.WARNING.value:
        logger.warning(f"{datetime.now()} {message}")

    elif log_level.value==LogLevel.ERROR.value:
        logger.error(f"{datetime.now()} {message}")

def parse_args():
    parser = argparse.ArgumentParser() # type: ignore

    parser.add_argument("--provider_id", help="candle_provider will go to work if from redis a matching topic partition_assign_topic with provider_id in it.", default=None)
    parser.add_argument("--market", help="Default BTC", default='BTC')
    parser.add_argument("--redis_ttl_ms", help="TTL for items published to redis. Default: 1000*60*60 (i.e. 1hr)",default=1000*60*60)

    args = parser.parse_args()
    if args.provider_id:
        param['provider_id'] = args.provider_id
    param['market'] = args.market
    param['redis_ttl_ms'] = int(args.redis_ttl_ms)


def init_redis_client() -> StrictRedis:
    redis_client : StrictRedis = StrictRedis(
                    host = param['mds']['redis']['host'],
                    port = param['mds']['redis']['port'],
                    db = 0,
                    ssl = False
                )
    try:
        redis_client.keys()
    except ConnectionError as redis_conn_error:
        err_msg = f"Failed to connect to redis: {param['mds']['redis']['host']}, port: {param['mds']['redis']['port']}"
        raise ConnectionError(err_msg)
    
    return redis_client


def _fetch_historical_daily_candle_height(
        exchange,
        normalized_symbol : str,
        timestamp_ms : int,
        offset_days : int,
        candle_height : float,
        reload_candle_height : bool = False
    ):
    if not candle_height or reload_candle_height:
        dt = datetime.fromtimestamp(int(timestamp_ms/1000)) + timedelta(days=offset_days)
        dt = datetime(dt.year, dt.month, dt.day)
        timestamp_ms = int(dt.timestamp()) * 1000
        if dt < datetime(datetime.today().year, datetime.today().month, datetime.today().day):
            historical_day_candle = fetch_ohlcv_one_candle(exchange=exchange, normalized_symbol=normalized_symbol, timestamp_ms=timestamp_ms, ref_timeframe='1d')
            if historical_day_candle:
                return historical_day_candle['close'] - historical_day_candle['open']
            else:
                return None
        else:
            return None
    else:
        return None

async def main():
    parse_args()
    
    param['job_name'] = f'candles_provider_{param["provider_id"]}'
    redis_client : StrictRedis = init_redis_client()

    exchange = deribit()

    i = 0
    while True:
        try:
            pd_old_expiry_data = None
            if os.path.isfile(param['archive_file_name']):
                pd_old_expiry_data = pd.read_csv(param['archive_file_name'])
                pd_old_expiry_data.drop(pd_old_expiry_data.columns[pd_old_expiry_data.columns.str.contains('unnamed',case = False)],axis = 1, inplace = True)
                pd_old_expiry_data['datetime'] = pd.to_datetime(pd_old_expiry_data['datetime'])
                pd_old_expiry_data['datetime'] = pd_old_expiry_data['datetime'].dt.tz_localize(None)

            start = time.time()
            expiry_data = fetch_deribit_btc_option_expiries(market = param['market'])
            expiry_data = expiry_data['by_expiry']
            elapsed_sec = int((time.time() - start))
            log(f"#{i} Took {elapsed_sec} sec to fetch option expiry data from Deribit")

            publish_key = param['mds']['topics']['deribit_options_expiry_publish_topic']
            expiry_sec = int(int(param['mds']['redis']['ttl_ms'])/1000)
            redis_client.set(name=publish_key, value=json.dumps(expiry_data).encode('utf-8'), ex=expiry_sec)

            pd_new_expiry_data = pd.DataFrame([ { 'datetime' : x[0], 'notional_usd' : x[1] } for x in expiry_data ])

            pd_new_expiry_data['symbol'] = f"{param['market']}/USDT"

            pd_new_expiry_data['datetime'] = pd.to_datetime(pd_new_expiry_data['datetime'])
            pd_new_expiry_data['datetime'] = pd_new_expiry_data['datetime'].dt.tz_localize(None)
            pd_new_expiry_data['timestamp_sec'] = pd_new_expiry_data['datetime'].apply(lambda dt: int(dt.timestamp()))
            pd_new_expiry_data['timestamp_ms'] = pd_new_expiry_data['timestamp_sec'] * 1000
            timestamp_to_datetime_cols(pd_new_expiry_data)

            if pd_old_expiry_data is not None:
                min_datetime_from_new_expiry_data = pd_new_expiry_data['datetime'].min()

                pd_old_expiry_data = pd_old_expiry_data[
                    pd_old_expiry_data['datetime'] < min_datetime_from_new_expiry_data # Take from new update, OI can change
                ]
                pd_merged_expiry_data = pd.concat([pd_old_expiry_data, pd_new_expiry_data], axis=0, ignore_index=True)

            else:
                pd_merged_expiry_data = pd_new_expiry_data

            if not 'daily_candle_height_tm0' in pd_merged_expiry_data.columns:
                pd_merged_expiry_data['daily_candle_height_tm0'] = None
                pd_merged_expiry_data['daily_candle_height_tm1'] = None
                pd_merged_expiry_data['daily_candle_height_tm2'] = None
                pd_merged_expiry_data['daily_candle_height_tm3'] = None

            # candle_height = close - open (Can be positive, can be negative)
            pd_merged_expiry_data['daily_candle_height_tm0'] = pd_merged_expiry_data.apply(lambda rw : _fetch_historical_daily_candle_height(exchange, rw['symbol'], rw['timestamp_ms'], 0, rw['daily_candle_height_tm0']), axis=1) # type: ignore
            pd_merged_expiry_data['daily_candle_height_tm1'] = pd_merged_expiry_data.apply(lambda rw : _fetch_historical_daily_candle_height(exchange, rw['symbol'], rw['timestamp_ms'], -1, rw['daily_candle_height_tm1']), axis=1) # type: ignore
            pd_merged_expiry_data['daily_candle_height_tm2'] = pd_merged_expiry_data.apply(lambda rw : _fetch_historical_daily_candle_height(exchange, rw['symbol'], rw['timestamp_ms'], -2, rw['daily_candle_height_tm2']), axis=1) # type: ignore
            pd_merged_expiry_data['daily_candle_height_tm3'] = pd_merged_expiry_data.apply(lambda rw : _fetch_historical_daily_candle_height(exchange, rw['symbol'], rw['timestamp_ms'], -3, rw['daily_candle_height_tm3']), axis=1) # type: ignore

            pd_merged_expiry_data.to_csv(param['archive_file_name'])

        except Exception as loop_err:
            log(f"Loop error {loop_err} {str(sys.exc_info()[0])} {str(sys.exc_info()[1])} {traceback.format_exc()}", log_level=LogLevel.ERROR)
        finally:
            i += 1

async def _run_jobs():
    await main()
asyncio.get_event_loop().run_until_complete(_run_jobs())


from enum import Enum
import argparse
import time
from datetime import datetime
from typing import Any, Dict, Union
import logging
import json
from tabulate import tabulate
import asyncio
import pandas as pd
import numpy as np
from redis import StrictRedis

'''
From command prompt:
    python aggregated_orderbook.py --normalized_symbol BTC/USDT:USDT --sliding_window_num_intervals 1200 --update_imabalce_csv_intervals 100 --dump_imbalance_to_disk Y --publish_imbalance_to_redis N
    or
    pypy aggregated_orderbook.py --normalized_symbol BTC/USDT:USDT --sliding_window_num_intervals 1200 --update_imabalce_csv_intervals 100 --dump_imbalance_to_disk Y --publish_imbalance_to_redis N

This script is pypy compatible.

Spot orderbooks REST API from exchanges
    Binance https://binance-docs.github.io/apidocs/spot/en/#order-book
    OKX https://www.okx.com/docs-v5/en/#order-book-trading-market-data-get-order-book
        'sz':  Order book depth per side. Maximum 400, e.g. 400 bids + 400 asks
    Bybit https://bybit-exchange.github.io/docs/v5/market/orderbook
    Coinbase https://docs.cdp.coinbase.com/exchange/reference/exchangerestapi_getproductbook
    Kraken https://docs.kraken.com/api/docs/rest-api/get-order-book/

Key parameters you may want to modify:
    normalized_symbol: Which ticker you wish monitor
    sliding_window_num_intervals: We calc EMAs on pct_imbalance, bids_amount_usdt and asks_amount_usdt. Window size will impact this. It's quoted in # intervals.
                                  It's measured in # intervals (or # loops), default 1200.
    update_imabalce_csv_intervals:  We'd update, or publish, imbalance data only at multiples of update_imabalce_csv_intervals.
                                  It's measured in # intervals (or # loops), default 100.
    topic_imbalance_data: Imbalance data is published to redis. This is published topic.
                           Since redis has special treatment for ':' (it'd consider it a folder), we'd replace ':' with '|'. 
                           Example BTC/USDT:USDT will become BTC/USDT|USDT.
    redis_ttl_ms: Imbalance data is published to redis. TTL of published data.
    dump_imbalance_to_disk: Dump imbalance data to disk?
    publish_imbalance_to_redis: Publish imbalance data to redis?

Launch.json if you wish to debug from VSCode:
    {
        "version": "0.2.0",
        "configurations": [
            {
                "name": "Python Debugger: Current File",
                "type": "debugpy",
                "request": "launch",
                "program": "${file}",
                "console": "integratedTerminal",
                "args" : [
                                    "--normalized_symbol", "BTC/USDT:USDT",
                                    "--sliding_window_num_intervals", "100",
                                    "--update_imabalce_csv_intervals", "100",
                                    "--topic_imbalance_data", "imbalance_BTC/USDT:USDT",
                                    "--redis_ttl_ms", "3600000",

                                    "--dump_imbalance_to_disk", "Y",
                                    "--publish_imbalance_to_redis", "N"
                ]
            }
        ]
    }

'''
from ccxt.binance import binance
from ccxt.okx import okx
from ccxt.bybit import bybit
from ccxt.coinbase import coinbase
from ccxt.kraken import kraken
from ccxt.base.exchange import Exchange

param : Dict = {
    'normalized_symbol' : 'BTC/USDT:USDT',
    'market_type' : 'linear', # For spots, set to "spot". For perpectual, you need to look at ccxt doc, for most exchanges, it's 'linear' or 'swap' for perpetuals. Example, https://github.com/ccxt/ccxt/blob/master/python/ccxt/okx.py?plain=1#L1110
    'depth' : 1000,
    'price_level_increment' : 10,
    'sliding_window_num_intervals' : 1200, # For example if each iteration takes 2 sec. 90 intervals = 180 sec (i.e. three minutes)
    'update_imabalce_csv_intervals' : 100,
    'imbalance_output_file' : 'imbalance.csv',
    'dump_imbalance_to_disk' : True,
    'publish_imbalance_to_redis' : False,

    # Provider ID is part of mds publish topic. 
    'provider_id' : 1,


    # Publish to message bus
    'mds' : {
        'mds_topic' : 'ccxt_rest_ob_$PROVIDER_ID$', 
        'redis' : {
            'host' : 'localhost',
            'port' : 6379,
            'db' : 0,
            'ttl_ms' : 1000*60*15 # 15 min?
        }

    },
    
    # Keep track of latency issues: ts_delta_observation_ms: Keep track of server clock vs timestamp from exchange
    'ts_delta_observation_ms_threshold' : 150
}

depth : int = param['depth']
market_type : str = param['market_type'] 
price_level_increment : float = param['price_level_increment']

logging.Formatter.converter = time.gmtime
logger = logging.getLogger()
log_level = logging.INFO # DEBUG --> INFO --> WARNING --> ERROR
logger.setLevel(log_level)
format_str = '%(asctime)s %(message)s'
formatter = logging.Formatter(format_str)
sh = logging.StreamHandler()
sh.setLevel(log_level)
sh.setFormatter(formatter)
logger.addHandler(sh)
# fh = logging.FileHandler(f"{param['job_name']}.log")
# fh.setLevel(log_level)
# fh.setFormatter(formatter)     
# logger.addHandler(fh)


coinbase_param : Dict[str, int] = { }

binance_param : Dict[str, int] = {
    'limit' : depth if depth <=5000 else 5000
}

okx_param : Dict[str, int] = {
    'sz' : depth if depth <= 400 else 400
}


bybit_param : Dict[str, int] = {
    'limit' : depth if depth <= 200 else 200
}


kraken_param : Dict[str, int] = {
    'depth' : depth if depth <= 500 else 500
}


coinbase_exchange = coinbase({
    'defaultType' : market_type
})

binance_exchange = binance({
    'defaultType' : market_type
})

okx_exchange = okx({
    'defaultType' : market_type
})

bybit_exchange = bybit({
    'defaultType' : market_type
})

kraken_exchange = kraken({
    'defaultType' : market_type
})

class LogLevel(Enum):
    CRITICAL = 50
    ERROR = 40
    WARNING = 30
    INFO = 20
    DEBUG = 10
    NOTSET = 0


def log(message : str, log_level : LogLevel = LogLevel.INFO):
    if log_level.value<LogLevel.WARNING.value:
        logger.info(f"{datetime.now()} {message}")

    elif log_level.value==LogLevel.WARNING.value:
        logger.warning(f"{datetime.now()} {message}")

    elif log_level.value==LogLevel.ERROR.value:
        logger.error(f"{datetime.now()} {message}")
    
async def _fetch_orderbook(symbol : str, exchange : Exchange, fetch_ob_params : Dict):
    try:
        ob = exchange.fetch_order_book(symbol=symbol, params=fetch_ob_params)
        is_valid = True
        ts_delta_observation_ms : int = 0
        if 'timestamp' in ob and ob['timestamp']:
            update_ts_ms = ob['timestamp']
            ts_delta_observation_ms = int(datetime.now().timestamp()*1000) - update_ts_ms
            is_valid = True if ts_delta_observation_ms<=param['ts_delta_observation_ms_threshold'] else False

        bid_prices = [ x[0] for x in ob['bids'] ]
        ask_prices = [ x[0] for x in ob['asks'] ]
        min_bid_price = min(bid_prices)
        max_bid_price = max(bid_prices)
        min_ask_price = min(ask_prices)
        max_ask_price = max(ask_prices)

        mid = (max([ x[0] for x in ob['bids'] ]) + min([ x[0] for x in ob['asks'] ])) / 2

        log(f"{exchange.name} mid: {mid}, min_bid_price: {min_bid_price}, max_bid_price: {max_bid_price}, min_ask_price: {min_ask_price}, max_ask_price: {max_ask_price}, range: {int(max_ask_price-min_bid_price)}, ts_delta_observation_ms: {ts_delta_observation_ms}")

        return {
                    'source' : exchange.name,
                    'orderbook' : ob,
                    'mid' : mid,
                    'min_bid_price' : min_bid_price,
                    'max_bid_price' : max_bid_price,
                    'min_ask_price' : min_ask_price,
                    'max_ask_price' : max_ask_price,
                    'is_valid' : is_valid,
                    'ts_delta_observation_ms' : ts_delta_observation_ms
                }
    except Exception as fetch_err:
        print(f"_fetch_orderbook failed for {exchange.name}: {fetch_err}")
        return {
            'source' : exchange.name,
            'is_valid' : False
        }
 

async def main():
    def parse_args():
        parser = argparse.ArgumentParser() # type: ignore

        parser.add_argument("--normalized_symbol", help="Example BTC/USDT for spot. BTC/USDT:USDT for perps.",default="BTC/USDT")
        parser.add_argument("--sliding_window_num_intervals", help="Sliding window is used for EMA's calculation. It's measured in # intervals (or # loops)",default=1200)
        parser.add_argument("--update_imabalce_csv_intervals", help="We'd update, or publish, imbalance data only at multiples of update_imabalce_csv_intervals. Again, it's measured in # intervals.",default=100)

        parser.add_argument("--topic_imbalance_data", help="Publish topic for imbalance data. Since redis has special treatment for ':' (it'd consider it a folder), we'd replace ':' with '|'. Example BTC/USDT:USDT will become BTC/USDT|USDT.",default=None)
        parser.add_argument("--redis_ttl_ms", help="TTL for items published to redis. Default: 1000*60*60 (i.e. 1hr)",default=1000*60*60)

        parser.add_argument("--dump_imbalance_to_disk", help="Y or N (default).", default='Y')
        parser.add_argument("--publish_imbalance_to_redis", help="Y or N (default).", default='N')
        
        args = parser.parse_args()
        param['normalized_symbol'] = args.normalized_symbol
        param['sliding_window_num_intervals'] = int(args.sliding_window_num_intervals)
        param['update_imabalce_csv_intervals'] = int(args.update_imabalce_csv_intervals)
        param['mds']['mds_topic'] = args.topic_imbalance_data if args.topic_imbalance_data else f"imbalance_data_{param['normalized_symbol'].replace(':','|')}"
        param['redis_ttl_ms'] = int(args.redis_ttl_ms)

        if args.dump_imbalance_to_disk:
            if args.dump_imbalance_to_disk=='Y':
                param['dump_imbalance_to_disk'] = True
            else:
                param['dump_imbalance_to_disk'] = False
        else:
            param['dump_imbalance_to_disk'] = False

        if args.publish_imbalance_to_redis:
            if args.publish_imbalance_to_redis=='Y':
                param['publish_imbalance_to_redis'] = True
            else:
                param['publish_imbalance_to_redis'] = False
        else:
            param['publish_imbalance_to_redis'] = False

    def init_redis_client():
        redis_client = StrictRedis(
                        host = param['mds']['redis']['host'],
                        port = param['mds']['redis']['port'],
                        db = 0,
                        ssl = False
                    )
        try:
            redis_client.keys()
        except ConnectionError as redis_conn_error:
            err_msg = f"Failed to connect to redis: {param['mds']['redis']['host']}, port: {param['mds']['redis']['port']}"
            raise ConnectionError(err_msg)
        
        return redis_client
    
    parse_args()
    redis_client = init_redis_client()

    normalized_symbol : str = param['normalized_symbol']
    param['job_name'] = f'ccxt_agg_ob_provider_{normalized_symbol.replace(":","_").replace("/","-")}'
    sliding_window_num_intervals : int = param['sliding_window_num_intervals']
    update_imabalce_csv_intervals : int = param['update_imabalce_csv_intervals']

    pd_imbalances = pd.DataFrame(columns=['timestamp_ms', 'datetime', 'mid', 'imbalance', 'bids_amount_usdt', 'asks_amount_usdt', 'total_amount', 'pct_imbalance', 'ema_pct_imbalance', 'ema_bids_amount_usdt', 'ema_asks_amount_usdt'])
    mid : Union[float, None] = None
    imbalance : Union[float, None] = None
    pct_imbalance : Union[float, None] = None
    sum_bids_amount_usdt : Union[float, None] = None
    sum_asks_amount_usdt : Union[float, None] = None
    total_amount : Union[float, None] = None
    last_ema_pct_imbalance : Union[float, None] = None
    last_ema_bids_amount_usdt : Union[float, None] = None
    last_ema_asks_amount_usdt : Union[float, None] = None

    i = 0
    total_loop_elapsed_ms : int = 0
    while True:
        loop_start = time.time()

        try:
            # Comment out Coinbase and Kraken spot only exchanges.
            orderbooks = await asyncio.gather(
                # _fetch_orderbook(symbol=normalized_symbol, exchange=coinbase_exchange, fetch_ob_params=coinbase_param),
                _fetch_orderbook(symbol=normalized_symbol, exchange=binance_exchange, fetch_ob_params=binance_param),
                _fetch_orderbook(symbol=normalized_symbol, exchange=bybit_exchange, fetch_ob_params=bybit_param),
                # _fetch_orderbook(symbol=normalized_symbol, exchange=kraken_exchange, fetch_ob_params=kraken_param),
                _fetch_orderbook(symbol=normalized_symbol, exchange=okx_exchange, fetch_ob_params=okx_param)
                )
            valid_orderbooks = [ ob for ob in orderbooks if ob['is_valid'] ]
            invalid_orderbooks = [ ob for ob in orderbooks if not ob['is_valid'] ]
            invalid_orderbooks_names = " ".join([ ob['source'] for ob in invalid_orderbooks ] )

            max_min_bid_price = max([ ob['min_bid_price'] for ob in valid_orderbooks if ob])
            best_bid_price = max([ob['max_bid_price'] for ob in valid_orderbooks if ob])
            min_max_ask_price = min([ob['max_ask_price'] for ob in valid_orderbooks if ob])
            best_ask_price = min([ob['min_ask_price'] for ob in valid_orderbooks if ob])

            elapsed_ms = (time.time() - loop_start) * 1000
            logger.info(f"orderbooks fetch elapsed (ms): {elapsed_ms}, # orderbooks: {len(valid_orderbooks)}, max_min_bid_price: {max_min_bid_price}, min_max_ask_price: {min_max_ask_price}, best_bid_price: {best_bid_price}, best_ask_price: {best_ask_price}. Invalid books: {invalid_orderbooks_names}")

            aggregated_orderbooks = {
                'bids' : {},
                'asks' : {}
            }

            def round_to_nearest(price, increment):
                return round(price / increment) * increment

            mid = [ x['mid'] for x in valid_orderbooks if x['source']=='Binance'][0] # use Binance as mid reference
            for orderbook in valid_orderbooks:
                bids = orderbook['orderbook']['bids']
                asks = orderbook['orderbook']['asks']

                for bid in bids:
                    price = round_to_nearest(bid[0], price_level_increment)
                    amount = bid[1]
                    if bid[0] > max_min_bid_price:
                        existing_amount = 0
                        if price in aggregated_orderbooks['bids']:
                            existing_amount = aggregated_orderbooks['bids'][price]['amount']
                        amount_in_base_ccy = existing_amount + amount
                        amount_in_usdt = amount_in_base_ccy * mid
                        aggregated_orderbooks['bids'][price] = {
                            'price' : price,
                            'amount' : amount_in_base_ccy,
                            'amount_usdt' : amount_in_usdt
                        }

                for ask in asks:
                    price = round_to_nearest(ask[0], price_level_increment)
                    amount = ask[1]
                    if ask[0] < min_max_ask_price:
                        existing_amount = 0
                        if price in aggregated_orderbooks['asks']:
                            existing_amount = aggregated_orderbooks['asks'][price]['amount']
                        amount_in_base_ccy = existing_amount + amount
                        amount_in_usdt = amount_in_base_ccy * mid
                        aggregated_orderbooks['asks'][price] = {
                            'price' : price,
                            'amount' : amount_in_base_ccy,
                            'amount_usdt' : amount_in_usdt
                        }
            
            sorted_asks = dict(sorted(aggregated_orderbooks['asks'].items(), key=lambda item: item[0], reverse=True))
            sorted_bids = dict(sorted(aggregated_orderbooks['bids'].items(), key=lambda item: item[0], reverse=True))

            pd_aggregated_orderbooks_asks = pd.DataFrame(sorted_asks)
            pd_aggregated_orderbooks_bids = pd.DataFrame(sorted_bids)

            pd_aggregated_orderbooks_asks = pd_aggregated_orderbooks_asks.transpose()
            pd_aggregated_orderbooks_bids = pd_aggregated_orderbooks_bids.transpose()

            sum_asks_amount_usdt = pd.to_numeric(pd_aggregated_orderbooks_asks['amount_usdt']).sum()
            sum_bids_amount_usdt = pd.to_numeric(pd_aggregated_orderbooks_bids['amount_usdt']).sum()

            pd_aggregated_orderbooks_asks['str_amount_usdt'] = pd_aggregated_orderbooks_asks['amount_usdt'].apply(lambda x: f'{x:,.2f}')
            pd_aggregated_orderbooks_bids['str_amount_usdt'] = pd_aggregated_orderbooks_bids['amount_usdt'].apply(lambda x: f'{x:,.2f}')

            ask_resistance_price_level = pd_aggregated_orderbooks_asks['amount_usdt'].idxmax()
            bid_support_price_level = pd_aggregated_orderbooks_bids['amount_usdt'].idxmax()

            pd_aggregated_orderbooks_asks['is_max_amount_usdt'] = pd_aggregated_orderbooks_asks.index == ask_resistance_price_level
            pd_aggregated_orderbooks_bids['is_max_amount_usdt'] = pd_aggregated_orderbooks_bids.index == bid_support_price_level

            pd_aggregated_orderbooks_asks_ = pd_aggregated_orderbooks_asks[['price', 'amount', 'str_amount_usdt', 'is_max_amount_usdt']]
            pd_aggregated_orderbooks_asks_.rename(columns={'str_amount_usdt': 'amount_usdt'}, inplace=True)
            pd_aggregated_orderbooks_bids_ = pd_aggregated_orderbooks_bids[['price', 'amount', 'str_amount_usdt', 'is_max_amount_usdt']]
            pd_aggregated_orderbooks_bids_.rename(columns={'str_amount_usdt': 'amount_usdt'}, inplace=True)

            spread_bps = (best_ask_price-best_bid_price) / mid * 10000
            spread_bps = round(spread_bps, 0)
            imbalance = sum_bids_amount_usdt - sum_asks_amount_usdt if sum_bids_amount_usdt and sum_asks_amount_usdt else None
            total_amount = sum_bids_amount_usdt + sum_asks_amount_usdt if sum_bids_amount_usdt and sum_asks_amount_usdt else None
            pct_imbalance = (imbalance/total_amount) * 100 if imbalance and total_amount else None

            log(f"mid: {mid}, imbalance (bids - asks): {imbalance:,.0f}, pct_imbalance: {pct_imbalance:,.2f}, last_ema_pct_imbalance: {last_ema_pct_imbalance if last_ema_pct_imbalance else '--'}, spread_bps between bests: {spread_bps} (If < 0, arb opportunity). Range {max_min_bid_price} - {min_max_ask_price} (${int(min_max_ask_price-max_min_bid_price)})")
            
            if last_ema_bids_amount_usdt and last_ema_asks_amount_usdt:
                log(f"ema_bids_amount_usdt: {last_ema_bids_amount_usdt:,.0f}, ema_asks_amount_usdt: {last_ema_asks_amount_usdt:,.0f}")

            log(f"asks USD {sum_asks_amount_usdt:,.0f}, best: {best_ask_price:,.2f}")
            log(f"{tabulate(pd_aggregated_orderbooks_asks_.reset_index(drop=True), headers='keys', tablefmt='psql', colalign=('right', 'right', 'right'), showindex=False)}") # type: ignore Otherwise error: tabulate Argument of type "DataFrame" cannot be assigned to parameter "tabular_data" of type "Mapping[str, Iterable[Any]] | Iterable[Iterable[Any]]" in function "tabulate"

            log(f"bids USD {sum_bids_amount_usdt:,.0f}, best: {best_bid_price:,.2f}")
            log(f"{tabulate(pd_aggregated_orderbooks_bids_.reset_index(drop=True), headers='keys', tablefmt='psql', colalign=('right', 'right', 'right'), showindex=False)}") # type: ignore Otherwise error: tabulate Argument of type "DataFrame" cannot be assigned to parameter "tabular_data" of type "Mapping[str, Iterable[Any]] | Iterable[Iterable[Any]]" in function "tabulate"

        except Exception as loop_err:
            log(f"#{i} Error: {loop_err}")

        finally:
            this_loop_elapsed_ms : int = int((time.time()-loop_start)*1000)
            total_loop_elapsed_ms += this_loop_elapsed_ms
            avg_loop_elapsed_ms : int = int(total_loop_elapsed_ms / (i+1))
            sliding_window_num_sec : int = int(sliding_window_num_intervals*avg_loop_elapsed_ms/1000)
            log(f"#{i} this_loop_elapsed_ms: {this_loop_elapsed_ms}, avg_loop_elapsed_ms: {avg_loop_elapsed_ms}, sliding_window_num_intervals: {sliding_window_num_intervals}, sliding_window_num_sec: {sliding_window_num_sec}")
        
        pct_imbalance = imbalance/total_amount * 100 if imbalance and total_amount else None
        pd_imbalances.loc[i] = [ int(loop_start*1000), datetime.fromtimestamp(loop_start), mid, imbalance, sum_bids_amount_usdt, sum_asks_amount_usdt, total_amount, pct_imbalance, np.nan, np.nan, np.nan ]
        
        if i%update_imabalce_csv_intervals==0:
            if pd_imbalances.shape[0]>sliding_window_num_intervals:
                pd_imbalances['ema_pct_imbalance'] = pd_imbalances['pct_imbalance'].ewm(span=sliding_window_num_intervals, adjust=False).mean()
                pd_imbalances['ema_bids_amount_usdt'] = pd_imbalances['bids_amount_usdt'].ewm(span=sliding_window_num_intervals, adjust=False).mean()
                pd_imbalances['ema_asks_amount_usdt'] = pd_imbalances['asks_amount_usdt'].ewm(span=sliding_window_num_intervals, adjust=False).mean()
                last_ema_pct_imbalance = pd_imbalances['ema_pct_imbalance'].iloc[-1]
                last_ema_bids_amount_usdt = pd_imbalances['ema_bids_amount_usdt'].iloc[-1]
                last_ema_asks_amount_usdt = pd_imbalances['ema_asks_amount_usdt'].iloc[-1]
            
                data : Dict[str, Union[str, float, int, None]] = {
                    'normalized_symbol' : normalized_symbol,
                    'timestamp_ms' : int(datetime.now().timestamp() * 1000),
                    
                    'ema_pct_imbalance' : last_ema_pct_imbalance,
                    'ema_bids_amount_usdt' : last_ema_bids_amount_usdt,
                    'ema_asks_amount_usdt' : last_ema_asks_amount_usdt,

                    'avg_loop_elapsed_ms' : avg_loop_elapsed_ms,
                    'sliding_window_num_intervals' : sliding_window_num_intervals,
                    'sliding_window_num_sec' : sliding_window_num_sec
                }
                redis_client.set(name=param['mds']['mds_topic'], value=json.dumps(data), ex=int(param['mds']['redis']['ttl_ms']/1000))
                
                if param['dump_imbalance_to_disk']:
                    pd_imbalances.to_csv(param['imbalance_output_file'])
            
        i += 1


asyncio.run(main())


from typing import Union, Dict

'''
https://gist.github.com/raphaelgabbarelli/bc5a41d93789046f9c71e4685f1463e7
https://www.youtube.com/watch?v=CFEYKrP0vxs
'''
import boto3
import base64

class AwsKmsUtil(object):
    def __init__(
                    self, 
                    key_id : str,
                    profile_name : Union[str,None] = None
                ):
        self.key_id = key_id
        aws_session = boto3.session.Session(profile_name=profile_name) # type: ignore "session" is not a known attribute of module "boto3"
        self.aws_kms_client = aws_session.client('kms')
    
    def encrypt(self, plaintext : str) -> bytes:
        encrypted = self.aws_kms_client.encrypt(KeyId=self.key_id, Plaintext=plaintext)
        encrypted : bytes = base64.b64encode(encrypted['CiphertextBlob']) # type: ignore
        return encrypted

    def decrypt(self, encrypted : bytes) -> str:
        decrypted : str = self.aws_kms_client.decrypt(CiphertextBlob=base64.b64decode(encrypted)) # type: ignore Cannot access attribute "ams_kms_client" for class "AwsKmsUtil*
        return decrypted['Plaintext'].decode('utf-8') # type: ignore

if __name__ == "__main__":
    '''
    From command line, run 'aws configure' with IAM user's Access key ID and Secret access key. (Assume you have awscli installed)
        aws configure
        AWS Access Key ID [****************ABCD]: <-- ***ABCD is your IAM user 
        AWS Secret Access Key [****************xxx]: <-- xxx is password to your IAM user
        Default region name [us-east-1]: <-- Region need be where your KMS key resides!
        Default output format [None]: 
    
    Remember that when you create your KMS Key, you need to grant permission of the key newly created key to IAM user (This is done on KMS side, not IAM).
    '''
    key_id : str = "" # Enter your KMS key ID here. You'd find it from under AWS > KMS > Customer managed keys
    original : str = "some secret"
    
    aws_kms = AwsKmsUtil(key_id=key_id, profile_name=None)
    encrypted = aws_kms.encrypt(original)
    decrpted = aws_kms.decrypt(encrypted)

    print(f"original: {original}, encrypted: {encrypted.decode('utf-8')}, decrpted: {decrpted}") # type: ignore


from enum import Enum
import argparse
import time
from datetime import datetime
from typing import Any, Dict, List, Union
import logging
import json
import pandas as pd
import numpy as np
from redis import StrictRedis

from ordergateway.client import DivisiblePosition

'''
set PYTHONPATH=%PYTHONPATH%;D:\dev\siglab\siglab_py
python test_ordergateway.py --gateway_id bybit_01
'''

param : Dict[str, str] = {
    'gateway_id' : '---'
}

def parse_args():
    parser = argparse.ArgumentParser() # type: ignore
    parser.add_argument("--gateway_id", help="gateway_id: Where are you sending your order?", default=None)

    args = parser.parse_args()
    param['gateway_id'] = args.gateway_id

def init_redis_client() -> StrictRedis:
    redis_client : StrictRedis = StrictRedis(
                    host = 'localhost',
                    port = 6379,
                    db = 0,
                    ssl = False
                )
    try:
        redis_client.keys()
    except ConnectionError as redis_conn_error:
        err_msg = f"Failed to connect to redis: {'localhost'}, port: {6379}"
        raise ConnectionError(err_msg)
    
    return redis_client

def execute_positions(
    redis_client : StrictRedis, 
    positions : List[DivisiblePosition], 
    ordergateway_pending_orders_topic : str):
    # https://redis.io/commands/publish/
    _positions = [ position.to_dict() for position in positions ]
    redis_client.set(name=ordergateway_pending_orders_topic, value=json.dumps(_positions).encode('utf-8'), ex=60*15)

    print(f"{ordergateway_pending_orders_topic}: Orders sent {_positions}.")

if __name__ == '__main__':
    parse_args()

    gateway_id : str = param['gateway_id']
    ordergateway_pending_orders_topic = 'ordergateway_pending_orders_$GATEWAY_ID$'
    ordergateway_pending_orders_topic = ordergateway_pending_orders_topic.replace("$GATEWAY_ID$", gateway_id)
    
    ordergateway_executions_topic = "ordergateway_executions_$GATEWAY_ID$"
    ordergateway_executions_topic = ordergateway_executions_topic.replace("$GATEWAY_ID$", gateway_id)

    redis_client : StrictRedis = init_redis_client()

    # Example, enter into a pair position long SUSHI, short DYDX
    positions_1 : List[DivisiblePosition] = [
        DivisiblePosition(
            ticker = 'SUSHI/USDT:USDT',
            side = 'sell',
            amount = 10,
            leg_room_bps = 5,
            order_type = 'limit',
            slices=5,
            wait_fill_threshold_ms=15000
        ),
        DivisiblePosition(
            ticker = 'DYDX/USDT:USDT',
            side = 'buy',
            amount = 10,
            leg_room_bps = 5,
            order_type = 'limit',
            slices=5,
            wait_fill_threshold_ms=15000
        )
    ]

    positions_2 : List[DivisiblePosition] = [
        DivisiblePosition(
            ticker = 'SUSHI/USDT:USDT',
            side = 'buy',
            amount = 10,
            leg_room_bps = 5,
            order_type = 'limit',
            slices=5,
            wait_fill_threshold_ms=15000
        ),
    ]

    positions_3 : List[DivisiblePosition] = [
        DivisiblePosition(
            ticker = 'BTC/USDT:USDT',
            side = 'sell',
            amount = 0.01,
            leg_room_bps = 5,
            order_type = 'limit',
            slices=1,
            wait_fill_threshold_ms=60000
        )
    ]
    execute_positions(
        redis_client=redis_client,
        positions=positions_2,
        ordergateway_pending_orders_topic=ordergateway_pending_orders_topic)

    # Wait for fills
    fills_received : bool = False
    while not fills_received:
        try:
            keys = redis_client.keys()
            for key in keys:
                s_key : str = key.decode("utf-8")
                if s_key==ordergateway_executions_topic:
                    message = redis_client.get(key)
                    if message:
                        message = message.decode('utf-8')
                        positions = json.loads(message)

                        for position in positions:
                            print(f"{position['ticker']} {position['side']} amount: {position['amount']} leg_room_bps: {position['leg_room_bps']} slices: {position['slices']}, filled_amount: {position['filled_amount']}, average_cost: {position['average_cost']}, # executions: {len(position['executions'])}")
                        fills_received = True
                        break

        except Exception as loop_err:
            print(f"Oops {loop_err}")
        finally:
            time.sleep(1)

    


import sys
import traceback
from enum import Enum
import argparse
import time
from datetime import datetime, timedelta
import operator
from typing import Any, Dict, Union, Mapping
import logging
import json
from tabulate import tabulate
import asyncio
from threading import Thread
from collections import defaultdict
import pandas as pd
import numpy as np
from redis import StrictRedis
from redis.client import PubSub

from ccxt.base.exchange import Exchange
import ccxt.pro as ccxtpro 

'''
To start from command prompt:
    set PYTHONPATH=%PYTHONPATH%;D:\dev\siglab
    python orderbooks_provider.py --provider_id aaa --instance_capacity 25 --ts_delta_observation_ms_threshold 150 --ts_delta_consecutive_ms_threshold 150 --redis_ttl_ms 3600000

This script is pypy compatible.

Key parameters you may want to modify:
    provider_id: You can trigger this provider instance using test_provider.py. Of course, you'd write your own.
    instance_capacity: max # tickers this provider instance will handle.
    ts_delta_observation_ms_threshold: default to 150ms. "Observation Delta" is clock diff between orderbook timestamp, and your local server clock.
    ts_delta_consecutive_ms_threshold: default to 150ms. "Consecutive Delta" is time elapsed between consecutive orderbook updates.
    redis_ttl_ms: This is how long orderbook snapshot will last on redis when provider publishes to it.

Launch.json if you wish to debug from VSCode:
    {
        "version": "0.2.0",
        "configurations": [
            {
                "name": "Python Debugger: Current File",
                "type": "debugpy",
                "request": "launch",
                "program": "${file}",
                "console": "integratedTerminal",
                "args" : [
                            "--provider_id", "YourProviderNameHere",
                            "--instance_capacity", 25",
                            "--ts_delta_observation_ms_threshold", "150",
                            "--ts_delta_consecutive_ms_threshold","150",
                            "--redis_ttl_ms", "3600000"
                ],
                "env": {
                    "PYTHONPATH": "${workspaceFolder}"
                }
            }
        ]
    }
'''

class LogLevel(Enum):
    CRITICAL = 50
    ERROR = 40
    WARNING = 30
    INFO = 20
    DEBUG = 10
    NOTSET = 0

param : Dict = {
    'market_type' : 'linear', # For spots, set to "spot". For perpectual, you need to look at ccxt doc, for most exchanges, it's 'linear' or 'swap' for perpetuals. Example, https://github.com/ccxt/ccxt/blob/master/python/ccxt/okx.py?plain=1#L1110
    
    # Provider ID is part of mds publish topic. 
    'provider_id' : 'ceaafe1d-e320-44ec-a959-da73edb9c4b1',

    "instance_capacity" : "25",
    
    # Keep track of latency issues
    # a) ts_delta_observation_ms: Keep track of server clock vs timestamp from exchange
    # b) ts_delta_consecutive_ms: Keep track of gap between consecutive updates
    'ts_delta_observation_ms_threshold' : 150,
    'ts_delta_consecutive_ms_threshold' : 150,

    # Publish to message bus
    'mds' : {
        'topics' : {
            'partition_assign_topic' : 'mds_assign_$PROVIDER_ID$',
            'candles_publish_topic' : 'orderbooks_$SYMBOL$_$EXCHANGE$'
        },
        'redis' : {
            'host' : 'localhost',
            'port' : 6379,
            'db' : 0,
            'ttl_ms' : 1000*60*15 # 15 min?
        }
    }    
}

logging.Formatter.converter = time.gmtime
logger = logging.getLogger()
log_level = logging.INFO # DEBUG --> INFO --> WARNING --> ERROR
logger.setLevel(log_level)
format_str = '%(asctime)s %(message)s'
formatter = logging.Formatter(format_str)
sh = logging.StreamHandler()
sh.setLevel(log_level)
sh.setFormatter(formatter)
logger.addHandler(sh)
# fh = logging.FileHandler(f"{param['job_name']}.log")
# fh.setLevel(log_level)
# fh.setFormatter(formatter)     
# logger.addHandler(fh)

market_type : str = param['market_type'] 

exchange_params : Dict = {
        'newUpdates': False,
        'options' : {
            'defaultType' : 'swap' # spot, swap
        }
    }

async def instantiate_exhange(
        exchange_name : str,
        old_exchange : Union[Exchange, None]
        ) -> Exchange:
    if old_exchange:
        await old_exchange.close() # type: ignore Otherwise, Error:  Cannot access attribute "close" for class "Exchange  Attribute "close" is unknown
    if exchange_name==f"binance_{market_type}":
        exchange = ccxtpro.binance(exchange_params)
    elif exchange_name==f"okx_{market_type}":
        exchange = ccxtpro.okx(exchange_params)
    elif exchange_name==f"bybit_{market_type}":
        exchange = ccxtpro.bybit(exchange_params)
    else:
        exchange = ccxtpro.binance(exchange_params)
    exchange.name = exchange_name # type: ignore Otherwise, Error: Cannot assign to attribute "name" for class "binance" "str" is not assignable to "None"
    return exchange

def log(message : str, log_level : LogLevel = LogLevel.INFO):
    if log_level.value<LogLevel.WARNING.value:
        logger.info(f"{datetime.now()} {message}")

    elif log_level.value==LogLevel.WARNING.value:
        logger.warning(f"{datetime.now()} {message}")

    elif log_level.value==LogLevel.ERROR.value:
        logger.error(f"{datetime.now()} {message}")

def parse_args():
    parser = argparse.ArgumentParser() # type: ignore

    parser.add_argument("--provider_id", help="candle_provider will go to work if from redis a matching topic partition_assign_topic with provider_id in it.", default=None)
    parser.add_argument("--instance_capacity", help="Instance capacity in num of tickers it can process. -1: No limit.", default=-1)
    parser.add_argument("--ts_delta_observation_ms_threshold", help="max threshold in ms: server clock vs update timestamp",default=param['ts_delta_observation_ms_threshold'])
    parser.add_argument("--ts_delta_consecutive_ms_threshold", help="max threshold in ms: gap between consecutive updates",default=param['ts_delta_consecutive_ms_threshold'])

    parser.add_argument("--redis_ttl_ms", help="TTL for items published to redis. Default: 1000*60*60 (i.e. 1hr)",default=1000*60*60)

    args = parser.parse_args()
    if args.provider_id:
        param['provider_id'] = args.provider_id
    param['instance_capacity'] = int(args.instance_capacity)
    param['ts_delta_observation_ms_threshold'] = int(args.ts_delta_observation_ms_threshold)
    param['ts_delta_consecutive_ms_threshold'] = int(args.ts_delta_consecutive_ms_threshold)
    param['redis_ttl_ms'] = int(args.redis_ttl_ms)

def init_redis_client() -> StrictRedis:
    redis_client : StrictRedis = StrictRedis(
                    host = param['mds']['redis']['host'],
                    port = param['mds']['redis']['port'],
                    db = 0,
                    ssl = False
                )
    try:
        redis_client.keys()
    except ConnectionError as redis_conn_error:
        err_msg = f"Failed to connect to redis: {param['mds']['redis']['host']}, port: {param['mds']['redis']['port']}"
        raise ConnectionError(err_msg)
    
    return redis_client
    
def init_redis_channel_subscription(redis_client : StrictRedis, partition_assign_topic : str) -> PubSub:
    redis_client = init_redis_client()
    pubsub = redis_client.pubsub()
    pubsub.subscribe(partition_assign_topic)
    return pubsub


class OrderBook:
    def __init__(
            self, 
            ticker : str, 
            exchange_name : str
            ) -> None:
        self.ticker : str = ticker
        self.exchange_name : str = exchange_name
        self.bids : Dict = {}
        self.asks : Dict = {}

        self.last_timestamp_ms : Union[int,None] = None
        self.timestamp : Union[int,None] = None # order book update timestamp in sec
        self.timestamp_ms : Union[int,None] = None # order book update timestamp in ms
        self.ts_delta_consecutive_ms : int = 0
        self.ts_delta_observation_ms : int = 0
        self.is_valid : bool = True

    def update_book(
                self, 
                update : Mapping,
                param : dict
                    ) -> None:
        update_ts_ms = update['timestamp'] 
        if len(str(update_ts_ms))==10:
            update_ts_ms = update_ts_ms*1000
        self.last_timestamp_ms = self.timestamp_ms
        self.timestamp_ms = int(update_ts_ms)
        self.timestamp = int(self.timestamp_ms/1000)

        '''
        Keep track of latency issues
        a) ts_delta_observation_ms: Keep track of server clock vs timestamp from exchange
        b) ts_delta_consecutive_ms: Keep track of gap between consecutive updates
        '''
        self.ts_delta_observation_ms = int(datetime.now().timestamp()*1000) - self.timestamp_ms
        self.ts_delta_consecutive_ms = self.timestamp_ms - self.last_timestamp_ms if self.last_timestamp_ms else 0

        self.is_valid = True
        if self.ts_delta_observation_ms>param['ts_delta_observation_ms_threshold']:
            self.is_valid = False
        if self.ts_delta_consecutive_ms>param['ts_delta_consecutive_ms_threshold']:
            self.is_valid = False

        self.bids.update((float(bid[0]), float(bid[1])) for bid in update.get('bids', []))
        self.asks.update((float(ask[0]), float(ask[1])) for ask in update.get('asks', []))
        self.bids = { key:val for key,val in self.bids.items() if val!=0}
        self.asks = { key:val for key,val in self.asks.items() if val!=0}
        if self.bids and self.asks:
            best_ask = float(min(self.asks.keys()))
            best_bid = float(max(self.bids.keys()))
            if best_ask<best_bid:
                raise ValueError(f"{self.exchange_name} {self.ticker} best bid >= best ask!?! How?")
        self.bids = dict(sorted(self.bids.items(), reverse=True))
        self.asks = dict(sorted(self.asks.items()))
    
    def to_dict(self):
        bids = ([float(price), float(amount)] for price, amount in self.bids.items() if float(amount))
        asks = ([float(price), float(amount)] for price, amount in self.asks.items() if float(amount))
        data = {
            "denormalized_ticker" : self.ticker,
            "exchange_name" : self.exchange_name,
            "bids" : sorted(bids, key=operator.itemgetter(0), reverse=True),
            "asks" : sorted(asks, key=operator.itemgetter(0)),
            "timestamp" : self.timestamp, # in sec
            "timestamp_ms" : self.timestamp_ms, # in ms (timestamp in update from exchange)
            'ts_delta_observation_ms' : self.ts_delta_observation_ms,
            'ts_delta_consecutive_ms' : self.ts_delta_consecutive_ms,
            "is_valid" : self.is_valid
        }

        data['best_ask'] = min(data['asks'])
        data['best_bid'] = min(data['bids'])
        return data
        
class ThreadTask:
        def __init__(self) -> None:
            self.keep_running = True

def handle_ticker(
    exchange_name : str,
    ticker : str,
    candles_publish_topic : str,
    redis_client : StrictRedis,
    task : ThreadTask
):
    asyncio.run(_handle_ticker(
            exchange_name=exchange_name,
            ticker=ticker,
            candles_publish_topic=candles_publish_topic,
            redis_client=redis_client,
            task=task
        )
    )

async def _handle_ticker(
    exchange_name : str,
    ticker : str,
    candles_publish_topic : str,
    redis_client : StrictRedis,
    task : ThreadTask
    ):
    exchange = await instantiate_exhange(exchange_name=exchange_name, old_exchange=None)
    ob = OrderBook(ticker=ticker, exchange_name=exchange_name)
    candles_publish_topic = candles_publish_topic.replace("$SYMBOL$", ticker)
    candles_publish_topic = candles_publish_topic.replace("$EXCHANGE$", exchange_name)
    
    local_server_timestamp_ms = datetime.now().timestamp()*1000
    exchange_timestamp_ms = await exchange.fetch_time() # type: ignore Otherwise, Error: Cannot access attribute "fetch_time" for class "Coroutine[Any, Any, binance | okx | bybit]"
    timestamp_gap_ms = local_server_timestamp_ms - exchange_timestamp_ms
    log(f"{exchange_name} {ticker} local_server_timestamp_ms: {local_server_timestamp_ms} vs exchange_timestamp_ms: {exchange_timestamp_ms}. timestamp_gap_ms: {timestamp_gap_ms}")

    while task.keep_running:
        try:
            update = await exchange.watch_order_book(ticker) # type: ignore Otherwise, Error: Cannot access attribute "watch_order_book" for class "Coroutine[Any, Any, binance | okx | bybit]"
            ob.update_book(update=update, param=param)

            ob_dict = ob.to_dict()

            redis_client.set(name=candles_publish_topic, value=json.dumps(ob_dict), ex=int(param['mds']['redis']['ttl_ms']/1000))

            ob_dict.pop('bids')
            ob_dict.pop('asks')
            pd_ob = pd.DataFrame(ob_dict)
            log(f"{tabulate(pd_ob, headers=pd_ob.columns)}") # type: ignore Otherwise, Error: Argument of type "DataFrame" cannot be assigned to parameter "tabular_data" of type "Mapping[str, Iterable[Any]]

        except ValueError as update_error:
            log(f"Update error! {update_error}")
            exchange = await instantiate_exhange(exchange_name=exchange_name, old_exchange=exchange) # type: ignore Otherwise, Error: Argument of type "Coroutine[Any, Any, Exchange] | Exchange" cannot be assigned to parameter "old_exchange" of type "Exchange | None" in function "instantiate_exhange"
            ob = OrderBook(ticker=ticker, exchange_name=exchange_name)

    
async def main():
    parse_args()
    
    param['job_name'] = f'candles_provider_{param["provider_id"]}'

    redis_client : StrictRedis = init_redis_client()
    partition_assign_topic : str = param['mds']['topics']['partition_assign_topic']
    partition_assign_topic = partition_assign_topic.replace("$PROVIDER_ID$", param['provider_id'])
    candles_publish_topic : str = param['mds']['topics']['candles_publish_topic']
    redis_pubsub : PubSub = init_redis_channel_subscription(redis_client, partition_assign_topic)

    log(f"orderbooks_provider {param['provider_id']} started, waiting for trigger. (Can use test_provider.py to trigger it)")
    
    tasks = []
    for message in redis_pubsub.listen():
        if message['type'] == 'message' and message['channel'].decode()==partition_assign_topic:
            if tasks:
                for task in tasks:
                    task.keep_running = False # Everytime provider get triggered, it'd cancel previous tasks

            tickers = json.loads(message['data'].decode('utf-8'))
            tickers = [ { 'exchange' : x.split('|')[0], 'ticker' : x.split('|')[-1] } for x in tickers ]
            logger.info(f"{partition_assign_topic} {message}")

            ticker_count : int = 1
            for entry in tickers:
                exchange_name : str = entry['exchange']
                ticker : str = entry['ticker']
                task : ThreadTask = ThreadTask()
                t = Thread(target=handle_ticker, args = (exchange_name, ticker, candles_publish_topic, redis_client, task))
                t.start()
                log(f"Task created for {exchange_name}, {ticker}")

                if ticker_count>=param['instance_capacity']:
                    break
                ticker_count+=1

asyncio.run(main())